////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//  @title    Buildable Vehicle                                               //
//  @author   Truce                                                           //
//  @version  1.0 - A New Hope                                                //
//                                                                            //
//  A vehicle consisting of groups of nodes that can be unhidden and colored  //
//  to visually represent bricks that players attempt to build on the chasis  //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
//  Warning: "chassis" is spelled as "chasis" in this code. Oops.

%error = ForceRequiredAddOn("Vehicle_Flying_Wheeled_Jeep");

if(%error == $Error::AddOn_Disabled)
	FlyingWheeledJeepVehicle.uiName = "";

else if(%error == $Error::AddOn_NotFound)
{
	error("ERROR: Vehicle_FlyingBuildableVehicle - required add-on Vehicle_Flying_Wheeled_Jeep not found");
	return;
}

$FlyingBuildableVehicle::Mode    = 1;                   // 6x144 seems to be less laggy
$FlyingBuildableVehicle::Mode[0] = "_org 4 6 6 6";      // 4 grids, 216 nodes
$FlyingBuildableVehicle::Mode[1] = "_alt 6 6 2 12";     // 6 grids, 144 nodes
$Server::Quota::Player     = $Max::Quota::Player; // use discretion, don't spam

datablock WheeledVehicleData(FlyingBuildableVehicle : FlyingWheeledJeepVehicle)
{
	numMountPoints = getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],1);
	shapeFile      = "./chasis" @ getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],0) @ ".dts";
	uiName         = "Flying Buildable Vehicle ";
};

datablock PlayerData(FlyingBuildableVehicleGrid : PlayerStandardArmor)
{
	shapeFile = "./grid" @ getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],0) @ ".dts";
	uiName    = "";
};

package FlyingBuildableVehicle
{
	function FlyingBuildableVehicle::onAdd(%this,%obj)
	{
		Parent::onAdd(%this,%obj);
		
		%mods = new ScriptGroup();
		%num  = getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],1);
		
		for(%i = 0; %i < %num; %i++)
		{
			%grid = new AIPlayer()
			{
				datablock = FlyingBuildableVehicleGrid;
			};
			
			if(!isObject(%grid))
			{
				%error++;
				continue;
			}
			
			%grid.hideNode("ALL");
			
			MissionCleanup.add(%grid);
			%obj.mountObject(%grid,%i + 1);
			%obj.FlyingBuildableVehicleGrid[%i] = %grid;
		}
		
		if(%error)
			error("ERROR: Vehicle_FlyingBuildableVehicle - Player quota hit, " @ %error @ " grids not created.");
		
		%obj.mods = %mods;
		MissionCleanup.add(%mods);
	}
	
	function FlyingBuildableVehicle::onRemove(%this,%obj)
	{
		%mods = %obj.mods;
		%num  = getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],1);
		
		for(%i = 0; %i < %num; %i++)
		{
			%grid = %obj.FlyingBuildableVehicleGrid[%i];
			
			if(isObject(%grid))
				%grid.schedule(0,delete);
		}
		
		if(isObject(%mods))
			%mods.schedule(0,delete);
	}
	
	function Vehicle::onActivate(%this,%obj,%cl,%pos,%idk)
	{
		if(%this.getDatablock() != FlyingBuildableVehicle.getID())
		{
			Parent::onActivate(%this,%obj,%cl,%pos,%idk);
			return;
		}
		
		if(%obj.FlyingBuildableVehicleMode != %this)
		{
			messageClient(%cl,'',"\c2You are \c3building \c2on this vehicle. Click again or die to stop.");
			%obj.FlyingBuildableVehicleMode  = %this;
			%obj.FlyingBuildableVehiclePaint = 1;
		}
		else
		{
			messageClient(%cl,'',"\c2You are \c0no longer building \c2on this vehicle.");
			%obj.FlyingBuildableVehicleMode = "";
		}
	}
	
	function serverCmdPlantBrick(%cl)
	{
		%pl    = %cl.player;
		%obj   = %pl.FlyingBuildableVehicleMode;
		%brick = %pl.tempBrick;
		%group = %obj.brickGroup;
		
		if(!isObject(%obj) || !isObject(%brick))
		{
			Parent::serverCmdPlantBrick(%cl);
			return;
		}
		
		if(getTrustLevel(%cl,%obj) < 2)
		{
			commandToClient(%cl,'centerPrint',%group.name @ " does not trust you enough to do that.",1);
			return;
		}
		
		// Find brick's offset to center of vehicle
		%posa  = %obj.getPosition();
		%posb  = %brick.getPosition();
		
		%distx = getWord(%posb,0) - getWord(%posa,0);
		%disty = getWord(%posb,1) - getWord(%posa,1);
		%distz = getWord(%posb,2) - getWord(%posa,2);
		
		// Adjust offset to compensate for direction
		%fv = %obj.getForwardVector();
		%vx = getWord(%fv,0);
		%vy = getWord(%fv,1);
		
		%offx = %distx * %vy + %disty * -%vx;
		%offy = %disty * %vy - %distx * -%vx;
		%offz = %distz;
		
		// Round and adjust values to offset origin
		%offx = mFloatLength((mFloatLength(%offx,2) + 1.25) * 2,0);
		%offy = mFloatLength((mFloatLength(%offy,2) + 2.75) * 2,0);
		%offz = mFloatLength((mFloatLength(%offz,2) - 0.20) * 5,0);
		
		// Compensate for vehicle and brick angles
		%data  = %brick.getDatablock();
		%sizex = %data.brickSizeX;
		%sizey = %data.brickSizeY;
		%sizez = %data.brickSizeZ;
		%rot   = mRadToDeg(mATan(%vx,%vy));
		
		if(%rot < 0)
			%rot += 360;
		
		     if(%rot >= 315 || %rot <  45) %angle = 3;
		else if(%rot >=  45 && %rot < 135) %angle = 0;
		else if(%rot >= 135 && %rot < 225) %angle = 1;
		else if(%rot >= 225 && %rot < 315) %angle = 2;
		
		if(angleID(%brick.angleID - %angle) % 2 == 0)
		{
			%temp  = %sizex;
			%sizex = %sizey;
			%sizey = %temp;
		}
		
		// Calculate the min/max of nodes to utilize
		%minx = %offx - (%sizex / 2 - 0.5) - 0.5 * !(%sizex % 2);
		%maxx = %offx + (%sizex / 2 + 0.5) - 0.5 * !(%sizex % 2);
		%miny = %offy - (%sizey / 2 - 0.5) - 0.5 * !(%sizey % 2);
		%maxy = %offy + (%sizey / 2 + 0.5) - 0.5 * !(%sizey % 2);
		%minz = %offz - (%sizez / 2 - 0.5) - 0.5 * !(%sizez % 2);
		%maxz = %offz + (%sizez / 2 + 0.5) - 0.5 * !(%sizez % 2);
		
		if(%minx < 0 || %maxx > 6 || %miny < 0 || %maxy > 12 || %minz < 0 || %maxz > 12)
		{
			commandToClient(%cl,'centerPrint',"You must place the brick entirely on the chassis.",1);
			return;
		}
		
		// Go through the nodes to unhide and color
		%mod = new ScriptObject();
		%col = getColorIDTable(%cl.currentColor);
		%pnt = %pl.FlyingBuildableVehiclePaint;
		
		%mode = $FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode];
		%cap  = getWord(%mode,1);
		%dimx = getWord(%mode,2);
		%dimy = getWord(%mode,3);
		%dimz = getWord(%mode,4);
		
		for(%z = %minz; %z < %maxz; %z++)
		{
			for(%y = %miny; %y < %maxy; %y++)
			{
				// Calculate the group of nodes to work with
				if(%cap == 4)
				{
					switch((%z > 5) * 2 + (%y > 5))
					{
						case 0: %q = %obj.FlyingBuildableVehicleGrid[1];
						case 1: %q = %obj.FlyingBuildableVehicleGrid[0];
						case 2: %q = %obj.FlyingBuildableVehicleGrid[3];
						case 3: %q = %obj.FlyingBuildableVehicleGrid[2];
					}
				}
				else
					%q = %obj.FlyingBuildableVehicleGrid[5 - mFloor(%y / 2)];
				
				// Calculate the name of the node being used
				%zz = %z % %dimz;
				%yy = %y % %dimy;
				
				for(%x = %minx; %x < %maxx; %x++)
				{
					%idx = %zz * %dimy * %dimx + %yy * %dimx + %x;
					
					if(strPos(%idx,".") != -1)
						continue;
					
					%node   = nodeName(%idx);
					%oldcol = %q.nodeColor[%node];
					%olddis = %q.nodeDisplay[%node];
					
					// Apply some actions to the specified node
					if(%pnt)
					{
						if(%oldcol !$= %col)
						{
							%mod.action[%i++]   = "COLOR " @ %q SPC %node SPC %oldcol;
							%q.nodeColor[%node] = %col;
							
							%q.setNodeColor(%node,%col);
						}
						
						if(!%olddis)
						{
							%mod.action[%i++]     = "HIDE " @ %q SPC %node;
							%q.nodeDisplay[%node] = 1;
							
							%q.unHideNode(%node);
						}
					}
					else
					{
						if(%olddis)
						{
							%mod.action[%i++]     = "UNHIDE " @ %q SPC %node;
							%q.nodeDisplay[%node] = 0;
							
							%q.hideNode(%node);
						}
					}
				}
			}
		}
		
		if(%i)
		{
			%cl.undoStack.push(%obj TAB "NODEMOD" TAB %mod);
			%obj.mods.add(%mod);
		}
		else
			%mod.schedule(0,delete);
	}
	
	function QueueSO::pop(%this)
	{
		%line = Parent::pop(%this);
		%type = getField(%line,1);
		
		if(%type !$= "NODEMOD")
			return %line;
		
		%mod  = getField(%line,2);
		%line = " ";
		
		if(!isObject(%mod))
			return;
		
		for(%i = 1; %line !$= ""; %i++)
		{
			%line = %mod.action[%i];
			%type = getWord(%line,0);
			%q    = getWord(%line,1);
			%node = getWord(%line,2);
			
			switch$(%type)
			{
				case "HIDE":
					%q.nodeDisplay[%node] = 0;
					%q.hideNode(%node);
				
				case "UNHIDE":
					%q.nodeDisplay[%node] = 1;
					%q.unHideNode(%node);
				
				case "COLOR":
					%col = getWords(%line,3);
					
					%q.nodeColor[%node] = %col;
					%q.setNodeColor(%node,%col);
			}
		}
		
		%mod.schedule(0,delete);
	}
	
	function paintProjectile::onCollision(%this,%obj,%col,%idk,%pos,%norm)
	{
		if(%col.getClassName() !$= "AIPlayer" || %col.getDatablock() != FlyingBuildableVehicleGrid.getID())
			Parent::onCollision(%this,%obj,%col,%idk,%pos,%norm);
	}
	
	function serverCmdUseSprayCan(%cl,%idx)
	{
		Parent::serverCmdUseSprayCan(%cl,%idx);
		
		%pl  = %cl.player;
		%obj = %pl.FlyingBuildableVehicleMode;
		%old = %pl.FlyingBuildableVehiclePaint;
		%idx = 1;
		
		if(!isObject(%obj) || %idx == %old)
			return;
		
		%pl.FlyingBuildableVehiclePaint = %idx;
		messageClient(%cl,'',"\c2You are now \c3placing \c2bricks when you build.");
	}
	
	function serverCmdUseFXCan(%cl,%idx)
	{
		Parent::serverCmdUseFXCan(%cl,%idx);
		
		%pl  = %cl.player;
		%obj = %pl.FlyingBuildableVehicleMode;
		%old = %pl.FlyingBuildableVehiclePaint;
		%idx = (%idx != 0);
		
		if(!isObject(%obj) || %idx == %old)
			return;
		
		%pl.FlyingBuildableVehiclePaint = %idx;
		
		if(!%idx && %old) messageClient(%cl,'',"\c2You are now \c0erasing \c2bricks when you build.");
		else              messageClient(%cl,'',"\c2You are now \c3placing \c2bricks when you build.");
	}
	
	function Vehicle::setScale(%this,%scale)
	{
		Parent::setScale(%this,%scale);
		
		if(%this.getDatablock() != FlyingBuildableVehicle.getID())
			return;
		
		%num = getWord($FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode],1);
		
		for(%i = 0; %i < %num; %i++)
		{
			%grid = %this.FlyingBuildableVehicleGrid[%i];
			
			if(isObject(%grid))
			{
				%grid.setScale(%scale);
				%this.mountObject(%grid,%i + 1);
			}
		}
	}
};
schedule(0,0,activatePackage,FlyingBuildableVehicle);

function serverCmdToggleFlyingChassis(%cl)
{
	// fml
	serverCmdToggleFlyingChasis(%cl);
}

function serverCmdToggleFlyingChasis(%cl)
{
	%pl    = %cl.player;
	%obj   = %pl.FlyingBuildableVehicleMode;
	%group = %obj.brickGroup;
	
	if(!isObject(%obj))
		return;
	
	if(getTrustLevel(%cl,%obj) < 2)
	{
		commandToClient(%cl,'centerPrint',%group.name @ " does not trust you enough to do that.",1);
		return;
	}
	
	%chasis = (%obj.chasisHide = !%obj.chasisHide);
	
	switch(%chasis)
	{
		case 0:
			%obj.unHideNode("ALL");
			messageClient(%cl,'',"\c2The chassis is now \c3unhidden\c2.");
		
		case 1:
			%obj.hideNode("ALL");
			messageClient(%cl,'',"\c2The chassis is now \c0hidden\c2.");
	}
}

function serverCmdSaveFlyingVehicle(%cl,%name)
{
	%pl    = %cl.player;
	%obj   = %pl.FlyingBuildableVehicleMode;
	%group = %obj.brickGroup;
	
	if(!isObject(%obj))
		return;
	
	if(getTrustLevel(%cl,%obj) < 2)
	{
		commandToClient(%cl,'centerPrint',%group.name @ " does not trust you enough to do that.",1);
		return;
	}
	
	if(!validateString(%name))
	{
		commandToClient(%cl,'centerPrint',"Invalid save file name: " @ %name,1);
		return;
	}
	
	%time = mFloor($Sim::Time - %cl.saveVehicleTime);
	
	if(!%cl.isSuperAdmin && %time < 5)
	{
		commandToClient(%cl,'centerPrint',"Please wait " @ 5 - %time @ " seconds before saving again.",1);
		return;
	}
	
	%so   = new ScriptObject();
	%args = $FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode];
	%cap  = getWord(%args,1);
	%dimx = getWord(%args,2);
	%dimy = getWord(%args,3);
	%dimz = getWord(%args,4);
	%num  = %dimx * %dimy * %dimz;
		
	for(%i = 0; %i < %cap; %i++)
	{
		%grid = %obj.FlyingBuildableVehicleGrid[%i];
		
		if(!isObject(%grid))
			continue;
		
		for(%j = 0; %j < %num; %j++)
		{
			%node  = nodeName(%j);
			%clean = strReplace(%node,".","dot");
			
			%so.nodeDisplay[%i,%clean] = %grid.nodeDisplay[%node];
			%so.nodeColor[%i,%clean]   = %grid.nodeColor[%node];
		}
	}
	
	%bl_id = ($Server::LAN) ? getNumKeyID() : %cl.bl_id;
	%path  = "config/server/vehicleSaves/" @ %bl_id @ "/" @ %name @ ".cs";
	
	%so.save(%path);
	%so.schedule(0,delete);
	
	%cl.saveVehicleTime = $Sim::Time;
	messageClient(%cl,'',"\c2You have \c3saved \c2the file: \c4" @ %name @ "\c2.");
}

function serverCmdLoadFlyingVehicle(%cl,%name)
{
	%pl    = %cl.player;
	%obj   = %pl.FlyingBuildableVehicleMode;
	%group = %obj.brickGroup;
	
	if(!isObject(%obj))
		return 1;	
	
	if(getTrustLevel(%cl,%obj) < 2)
	{
		commandToClient(%cl,'centerPrint',%group.name @ " does not trust you enough to do that.",1);
		return;
	}
	
	%bl_id = ($Server::LAN) ? getNumKeyID() : %cl.bl_id;
	%path  = "config/server/vehicleSaves/" @ %bl_id @ "/" @ %name @ ".cs";
	
	if(!validateString(%name) || !isFile(%path))
	{
		commandToClient(%cl,'centerPrint',"Invalid load file name: " @ %name,1);
		return;
	}
	
	%time = mFloor($Sim::Time - %cl.loadVehicleTime);
	
	if(!%cl.isSuperAdmin && %time < 30)
	{
		commandToClient(%cl,'centerPrint',"Please wait " @ 30 - %time @ " seconds before loading again.",1);
		return;
	}
	
	%data = "$FlyingBuildableVehicle::LoadSO = ";
	%file = new FileObject();
	
	%file.openForRead(%path);
	%file.readLine();
	
	while(!%file.isEOF())
		%data = %data @ %file.readLine() @ "\n";
	
	%file.close();
	%file.delete();
	
	eval(%data);
	
	%so   = $FlyingBuildableVehicle::LoadSO;
	%args = $FlyingBuildableVehicle::Mode[$FlyingBuildableVehicle::Mode];
	%cap  = getWord(%args,1);
	%dimx = getWord(%args,2);
	%dimy = getWord(%args,3);
	%dimz = getWord(%args,4);
	%num  = %dimx * %dimy * %dimz;
		
	for(%i = 0; %i < %cap; %i++)
	{
		%grid = %obj.FlyingBuildableVehicleGrid[%i];
		
		if(!isObject(%grid))
			continue;
		
		%grid.hideNode("ALL");
		
		for(%j = 0; %j < %num; %j++)
		{
			%node  = nodeName(%j);
			%clean = strReplace(%node,".","dot");
			
			%dis = %so.nodeDisplay[%i,%clean];
			%col = %so.nodeColor[%i,%clean];
			
			%grid.nodeDisplay[%node] = %dis;
			%grid.nodeColor[%node] = %col;
			
			if(%dis)
				%grid.unHideNode(%node);
			
			if(%col !$= "")
				%grid.setNodeColor(%node,%col);
		}
	}
	
	%so.schedule(0,delete);
	
	%cl.loadVehicleTime = $Sim::Time;
	messageClient(%cl,'',"\c2You have \c0loaded \c2the file: \c4" @ %name @ "\c2.");
}

function angleID(%id)
{
	while(%id < 0) %id += 4;
	while(%id > 3) %id -= 4;
	
	return %id;
}

function nodeName(%idx)
{
	%node = "Cube";
	
	if(%idx)
	{
		while(strLen(%idx) < 3)
			%idx = "0" @ %idx;
		
		%node = %node @ "." @ %idx;
	}
	
	return %node;
}

function validateString(%str)
{
	%chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-";
	%len   = strLen(%str);
	
	if(%len < 3 || %len > 12)
		return 0;
	
	for(%i = 0; %i < %len; %i++)
	{
		if(striPos(%chars,getSubStr(%str,%i,1)) == -1)
			return 0;
	}	
	return 1;
}